"""
Copyright 2017-2019 Government of Canada - Public Services and Procurement Canada - buyandsell.gc.ca

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""


import json
import logging

from ctypes import CDLL

from indy import wallet
from indy.error import IndyError, ErrorCode

from von_anchor.error import BadAccess, ExtantWallet, WalletState
from von_anchor.wallet import DIDInfo, Wallet


LOGGER = logging.getLogger(__name__)


class WalletManager:
    """
    Class encapsulating management functionality for indy-sdk wallets.
    """

    def __init__(self, defaults: dict = None) -> None:
        """
        Initializer for wallet manager. Store default values by dict key:

            - 'storage_type': storage type (default None)
            - 'freshness_time': freshness time (default indefinite)
            - 'auto_create': auto_create behaviour (default False)
            - 'auto_remove': auto_remove behaviour (default False)
            - 'key': access credentials value (default 'key', for indy wallet access credentials {'key': 'key'}).

        :param defaults: default values to use as above
        """

        LOGGER.debug('WalletManager.__init__ >>> defaults %s', defaults)

        self._defaults = {
            'storage_type': (defaults or {}).get('storage_type', None),
            'freshness_time': int((defaults or {}).get('freshness_time', 0)),
            'auto_create': bool((defaults or {}).get('auto_create', False)),
            'auto_remove': bool((defaults or {}).get('auto_remove', False)),
            'access_creds': {'key': (defaults or {}).get('key', 'key')}
        }

        LOGGER.debug('WalletManager.__init__ <<<')

    @property
    def default_storage_type(self) -> str:
        """
        Accessor for default wallet storage type.

        :return: default storage type
        """

        return self._defaults['storage_type']

    @property
    def default_freshness_time(self) -> int:
        """
        Accessor for default freshness time.

        :return: default freshness time (0 for indefinite)
        """

        return self._defaults['freshness_time']

    @property
    def default_auto_create(self) -> bool:
        """
        Accessor for default auto_create behaviour.

        :return: default auto_create behaviour
        """

        return self._defaults['auto_create']

    @property
    def default_auto_remove(self) -> bool:
        """
        Accessor for default auto_remove behaviour.

        :return: default auto_remove behaviour
        """

        return self._defaults['auto_remove']

    @property
    def default_access(self) -> dict:
        """
        Accessor for default access credential value.

        :return: default access credential value.
        """

        return self._defaults['access_creds']['key']

    def _config2indy(self, config: dict) -> dict:
        """
        Given a configuration dict with indy and possibly more configuration values, return the
        corresponding indy wallet configuration dict from current default and input values.

        :param config: input configuration
        :return: configuration dict for indy wallet
        """

        assert {'name', 'id'} & {k for k in config}
        return {
            'id': config.get('name', config.get('id')),
            'storage_type': config.get('storage_type', self.default_storage_type),
            'freshness_time': config.get('freshness_time', self.default_freshness_time)
        }

    def _config2von(self, config: dict, access: str = None) -> dict:
        """
        Given a configuration dict with indy and possibly more configuration values, return the
        corresponding VON wallet configuration dict from current default and input values.

        :param config: input configuration
        :param access: access credentials value
        :return: configuration dict for VON wallet with VON-specific entries
        """

        rv = {k: config.get(k, self._defaults[k]) for k in ('auto_create', 'auto_remove')}
        rv['access'] = access or self.default_access
        for key in ('seed', 'did', 'link_secret_label'):
            if key in config:
                rv[key] = config[key]
        return rv

    async def create(self, config: dict = None, access: str = None, replace: bool = False) -> Wallet:
        """
        Create wallet on input name with given configuration and access credential value.

        Raise ExtantWallet if wallet on input name exists already and replace parameter is False.
        Raise BadAccess on replacement for bad access credentials value.

        FAIR WARNING: specifying replace=True attempts to remove any matching wallet before proceeding; to
        succeed, the existing wallet must use the same access credentials that the input configuration has.

        :param config: configuration data for both indy-sdk and VON anchor wallet:

            - 'name' or 'id': wallet name
            - 'storage_type': storage type
            - 'freshness_time': freshness time
            - 'did': (optional) DID to use
            - 'seed': (optional) seed to use
            - 'auto_create': whether to create the wallet on first open (persists past close, can work with auto_remove)
            - 'auto_remove': whether to remove the wallet on next close
            - 'link_secret_label': (optional) link secret label to use to create link secret

        :param access: indy wallet access credential ('key') value, if different than default
        :param replace: whether to replace old wallet if it exists
        :return: wallet created
        """

        LOGGER.debug('WalletManager.create >>> config %s, access %s, replace %s', config, access, replace)

        assert {'name', 'id'} & {k for k in config}
        wallet_name = config.get('name', config.get('id'))
        if replace:
            von_wallet = self.get(config, access)
            if not await von_wallet.remove():
                LOGGER.debug('WalletManager.create <!< Failed to remove wallet %s for replacement', wallet_name)
                raise ExtantWallet('Failed to remove wallet {} for replacement'.format(wallet_name))

        indy_config = self._config2indy(config)
        von_config = self._config2von(config, access)
        rv = Wallet(indy_config, von_config)
        await rv.create()
        LOGGER.debug('WalletManager.create <<< %s', rv)
        return rv

    def get(self, config: dict, access: str = None) -> Wallet:
        """
        Instantiate and return VON anchor wallet object on given configuration, respecting wallet manager
        default configuration values.

        :param config: configuration data for both indy-sdk and VON anchor wallet:

            - 'name' or 'id': wallet name
            - 'storage_type': storage type
            - 'freshness_time': freshness time
            - 'did': (optional) DID to use
            - 'seed': (optional) seed to use
            - 'auto_create': whether to create the wallet on first open (persists past close, can work with auto_remove)
            - 'auto_remove': whether to remove the wallet on next close
            - 'link_secret_label': (optional) link secret label to use to create link secret

        :param access: indy access credentials value
        :return: VON anchor wallet
        """

        LOGGER.debug('WalletManager.get >>> config %s, access %s', config, access)

        rv = Wallet(
            self._config2indy(config),
            self._config2von(config, access))

        LOGGER.debug('WalletManager.get <<< %s', rv)
        return rv

    async def reseed_local(self, local_wallet: Wallet, next_seed: str = None) -> DIDInfo:
        """
        Generate and apply new key, in wallet only, for local DID based on input seed (default random).
        Raise WalletState if wallet is closed.

        Note that this operation does not update the corresponding NYM on the ledger: for VON anchors
        anchored to the ledger, use von_anchor.BaseAnchor.reseed().

        :param local_wallet: VON anchor wallet without NYM on ledger
        :param next_seed: incoming replacement seed (default random)
        :return: DIDInfo with new verification key and metadata for DID
        """

        LOGGER.debug('WalletManager.reseed_local >>> local_wallet %s', local_wallet)

        await local_wallet.reseed_init(next_seed)

        rv = await local_wallet.reseed_apply()
        LOGGER.debug('WalletManager.reseed_local <<< %s', rv)
        return rv

    async def export_wallet(self, von_wallet: Wallet, path: str) -> None:
        """
        Export an existing VON anchor wallet. Raise WalletState if wallet is closed.

        :param von_wallet: open wallet
        :param path: path to which to export wallet
        """

        LOGGER.debug('WalletManager.export_wallet >>> von_wallet %s, path %s', von_wallet, path)

        if not von_wallet.handle:
            LOGGER.debug('WalletManager.export_wallet <!< Wallet %s is closed', von_wallet.name)
            raise WalletState('Wallet {} is closed'.format(von_wallet.name))

        await wallet.export_wallet(
            von_wallet.handle,
            json.dumps({
                'path': path,
                **von_wallet.access_creds
            }))

        LOGGER.debug('WalletManager.export_wallet <<<')

    async def import_wallet(self, indy_config: dict, path: str, access: str = None) -> None:
        """
        Import a VON anchor wallet. Raise BadAccess on bad access credential value.

        :param indy_config: indy wallet configuration to use, with:

            - 'id'
            - 'storage_type' (optional)
            - 'storage_config' (optional)

        :param path: path from which to import wallet file
        :param access: indy access credentials value (default value from wallet manager)
        """

        LOGGER.debug('WalletManager.import_wallet >>> indy_config %s, path: %s', indy_config, path)

        try:
            await wallet.import_wallet(
                json.dumps(indy_config),
                json.dumps({'key': access or self.default_access}),
                json.dumps({'path': path, 'key': access or self.default_access}))
        except IndyError as x_indy:
            if x_indy.error_code == ErrorCode.CommonInvalidStructure:  # indy-sdk raises on bad access
                LOGGER.debug(
                    'WalletManager.import_wallet <!< bad access credential value for wallet %s',
                    indy_config.get('id', '(no id)'))
                raise BadAccess('Bad access credential value for wallet {}'.format(indy_config.get('id', '(no id)')))
            LOGGER.debug(
                'WalletManager.import_wallet <!< indy error code %s on wallet %s import',
                x_indy.error_code,
                indy_config.get('id', '(no id)'))
            raise

        LOGGER.debug('WalletManager.import_wallet <<<')

    async def reset(self, von_wallet: Wallet, seed: str = None) -> Wallet:
        """
        Close and delete (open) VON anchor wallet and then create, open, and return
        replacement on current link secret.

        Note that this operation effectively destroys private keys for keyed data
        structures such as credential offers or credential definitions.

        Raise WalletState if the wallet is closed.

        :param von_wallet: open wallet
        :param seed: seed to use for new wallet (default random)
        :return: replacement wallet
        """

        LOGGER.debug('WalletManager.reset >>> von_wallet %s', von_wallet)

        if not von_wallet.handle:
            LOGGER.debug('WalletManager.reset <!< Wallet %s is closed', von_wallet.name)
            raise WalletState('Wallet {} is closed'.format(von_wallet.name))

        w_config = von_wallet.config  # wallet under reset, no need to make copy
        w_config['did'] = von_wallet.did
        w_config['seed'] = seed
        w_config['auto_create'] = von_wallet.auto_create  # in case both auto_remove+auto_create set (create every open)
        w_config['auto_remove'] = von_wallet.auto_remove

        label = await von_wallet.get_link_secret_label()
        if label:
            w_config['link_secret_label'] = label

        await von_wallet.close()
        if not von_wallet.auto_remove:
            await self.remove(von_wallet)

        rv = await self.create(w_config, von_wallet.access)
        await rv.open()

        LOGGER.debug('WalletManager.reset <<< %s', rv)
        return rv

    async def remove(self, von_wallet: Wallet) -> None:
        """
        Remove serialized wallet if it exists. Raise WalletState if wallet is open.

        :param von_wallet: (closed) wallet to remove
        """

        LOGGER.debug('WalletManager.remove >>> wallet %s', von_wallet)

        await von_wallet.remove()

        LOGGER.debug('WalletManager.remove <<<')

    @staticmethod
    async def register_storage_library(storage_type: str, c_library: str, entry_point: str) -> None:
        """
        Load a wallet storage plug-in.

        An indy-sdk wallet storage plug-in is a shared library; relying parties must explicitly
        load it before creating or opening a wallet with the plug-in.

        The implementation loads a dynamic library and calls an entry point; internally,
        the plug-in calls the indy-sdk wallet
        async def register_wallet_storage_library(storage_type: str, c_library: str, fn_pfx: str).

        :param storage_type: wallet storage type
        :param c_library: plug-in library
        :param entry_point: function to initialize the library
        """

        LOGGER.debug(
            'WalletManager.register_storage_library >>> storage_type %s, c_library %s, entry_point %s',
            storage_type,
            c_library,
            entry_point)

        try:
            stg_lib = CDLL(c_library)
            result = stg_lib[entry_point]()
            if result:
                LOGGER.debug(
                    'WalletManager.register_storage_library <!< indy error code %s on storage library entry at %s',
                    result,
                    entry_point)
                raise IndyError(result)
            LOGGER.info('Loaded storage library type %s (%s)', storage_type, c_library)
        except IndyError as x_indy:
            LOGGER.debug(
                'WalletManager.register_storage_library <!< indy error code %s on load of storage library %s %s',
                x_indy.error_code,
                storage_type,
                c_library)
            raise

        LOGGER.debug('WalletManager.register_storage_library <<<')



"""
Copyright 2017-2019 Government of Canada - Public Services and Procurement Canada - buyandsell.gc.ca

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""


import asyncio
import json
import logging

from os import listdir, makedirs, remove, rename
from os.path import basename, isdir, isfile, join
from shutil import rmtree

from indy import anoncreds, ledger
from indy.error import IndyError, ErrorCode

from von_anchor.anchor.base import BaseAnchor
from von_anchor.anchor.rrbuilder import RevRegBuilder
from von_anchor.cache import RevoCacheEntry, CRED_DEF_CACHE, REVO_CACHE
from von_anchor.error import (
    AbsentCredDef,
    AbsentPool,
    AbsentSchema,
    AbsentRevReg,
    AbsentTails,
    BadIdentifier,
    BadLedgerTxn,
    BadRevocation,
    CorruptWallet,
    WalletState)
from von_anchor.indytween import cred_attr_value
from von_anchor.nodepool import NodePool
from von_anchor.tails import Tails
from von_anchor.util import (
    cred_def_id,
    cred_def_id2seq_no,
    ok_cred_def_id,
    ok_rev_reg_id,
    ok_schema_id,
    rev_reg_id,
    rev_reg_id2cred_def_id,
    rev_reg_id2cred_def_id_tag,
    schema_key)
from von_anchor.wallet import Wallet


LOGGER = logging.getLogger(__name__)


class Issuer(BaseAnchor):
    """
    Mixin for anchor acting in Issuer capacity. An Issuer creates credential definitions and
    sends them to the ledger, issues credentials, and revokes credentials. Revocation support
    involves the management of tails files and revocation registries.
    """

    def __init__(self, wallet: Wallet, pool: NodePool = None, **kwargs) -> None:
        """
        Initializer for issuer.
        """

        LOGGER.debug('Issuer.__init__ >>> wallet: %s, pool: %s, kwargs: %s', wallet, pool, kwargs)

        super().__init__(wallet, pool, **kwargs)
        self._rrb = RevRegBuilder(wallet, **kwargs)

        LOGGER.debug('Issuer.__init__ <<<')

    @property
    def rrb(self) -> RevRegBuilder:
        """
        Issuer's revocation registry builder

        :return: revocation registry builder (anchor)
        """

        return self._rrb

    @property
    def rrbx(self) -> bool:
        """
        Whether revocation registry builder posture is external.

        :return: True for external, False for internal
        """

        return self.rrb.external

    @property
    def dir_tails(self) -> str:
        """
        Root of tails directory.

        :return: tails directory root
        """

        return self.rrb.dir_tails

    async def open(self) -> 'Issuer':
        """
        Explicit entry. Perform ancestor opening operations,
        then synchronize revocation registry to tails tree content.

        :return: current object
        """

        LOGGER.debug('Issuer.open >>>')

        await super().open()
        for path_rr_id in Tails.links(self.dir_tails, self.did):
            await self._sync_revoc_for_issue(basename(path_rr_id))

        LOGGER.debug('Issuer.open <<<')
        return self

    async def _send_rev_reg_def(self, rr_id: str) -> None:
        """
        Move tails file from hopper; deserialize revocation registry definition and initial entry;
        send to ledger and cache revocation registry definition.

        Operation serializes to subdirectory within tails hopper directory; symbolic
        link presence signals completion.

        Raise AbsentRevReg if revocation registry is not ready in hopper, or AbsentTails if
        tails file is not yet linked by its revocation registry identifier.

        :param rr_id: revocation registry identifier
        """

        LOGGER.debug('Issuer._send_rev_reg_def >>> rr_id: %s', rr_id)

        dir_tails_rr_id = self.rrb.dir_tails_top(rr_id)
        dir_target = self.rrb.dir_tails_target(rr_id)

        if not Tails.linked(dir_tails_rr_id, rr_id):
            LOGGER.debug(
                'Issuer._send_rev_reg_def <!< Tails file for rev reg %s not ready in dir %s',
                rr_id,
                dir_target)
            raise AbsentRevReg('Tails file for rev reg {} not ready in dir {}'.format(rr_id, dir_target))

        file_rr_def = join(dir_target, 'rr_def.json')
        if not isfile(file_rr_def):
            LOGGER.debug('Issuer._send_rev_reg_def <!< Rev reg def file %s not present', file_rr_def)
            raise AbsentRevReg('Rev reg def file {} not present'.format(file_rr_def))
        with open(file_rr_def, 'r') as fh_rr_def:
            rr_def_json = fh_rr_def.read()

        file_rr_ent = join(dir_target, 'rr_ent.json')
        if not isfile(file_rr_ent):
            LOGGER.debug('Issuer._send_rev_reg_def <!< Rev reg entry file %s not present', file_rr_ent)
            raise AbsentRevReg('Rev reg entry file {} not present'.format(file_rr_ent))
        with open(file_rr_ent, 'r') as fh_rr_ent:
            rr_ent_json = fh_rr_ent.read()

        file_tails = Tails.linked(dir_tails_rr_id, rr_id)
        if not file_tails:
            LOGGER.debug('Issuer._send_rev_reg_def <!< Tails link %s not present in dir %s', rr_id, dir_target)
            raise AbsentTails('Tails link {} not present in dir {}'.format(rr_id, dir_target))

        if self.rrbx:
            dir_cd_id = join(self.dir_tails, rev_reg_id2cred_def_id(rr_id))
            makedirs(dir_cd_id, exist_ok=True)
            rename(file_tails, join(dir_cd_id, basename(file_tails)))

        with REVO_CACHE.lock:
            rr_def_req_json = await ledger.build_revoc_reg_def_request(self.did, rr_def_json)
            await self._sign_submit(rr_def_req_json)
            await self.get_rev_reg_def(rr_id)  # add to cache en passant

        rr_ent_req_json = await ledger.build_revoc_reg_entry_request(self.did, rr_id, 'CL_ACCUM', rr_ent_json)
        await self._sign_submit(rr_ent_req_json)

        if self.rrbx:
            Tails.associate(self.dir_tails, rr_id, basename(file_tails))
            rmtree(dir_tails_rr_id)
        else:
            remove(file_rr_def)
            remove(file_rr_ent)

        LOGGER.debug('Issuer._send_rev_reg_def <<<')

    async def _set_rev_reg(self, rr_id: str, rr_size: int) -> None:
        """
        Move precomputed revocation registry data from hopper into place within tails directory.

        :param rr_id: revocation registry identifier
        :param rr_size: revocation registry size, in case creation required
        """

        LOGGER.debug('Issuer._set_rev_reg >>> rr_id: %s, rr_size: %s', rr_id, rr_size)
        assert self.rrbx

        dir_hopper_rr_id = join(self.rrb.dir_tails_hopper, rr_id)

        while Tails.linked(dir_hopper_rr_id, rr_id) is None:
            await asyncio.sleep(1)
        await self._send_rev_reg_def(rr_id)

        cd_id = rev_reg_id2cred_def_id(rr_id)
        (next_tag, rr_size_suggested) = Tails.next_tag(self.dir_tails, cd_id)
        rr_id = rev_reg_id(cd_id, next_tag)
        self.rrb.mark_in_progress(rr_id, rr_size or rr_size_suggested)

        LOGGER.debug('Issuer._set_rev_reg <<<')

    async def _sync_revoc_for_issue(self, rr_id: str, rr_size: int = None) -> None:
        """
        Create revocation registry if need be for input revocation registry identifier;
        open and cache tails file reader.

        :param rr_id: revocation registry identifier
        :param rr_size: if new revocation registry necessary, its size (default as per RevRegBuilder.create_rev_reg())
        """

        LOGGER.debug('Issuer._sync_revoc_for_issue >>> rr_id: %s, rr_size: %s', rr_id, rr_size)

        if not ok_rev_reg_id(rr_id):
            LOGGER.debug('Issuer._sync_revoc_for_issue <!< Bad rev reg id %s', rr_id)
            raise BadIdentifier('Bad rev reg id {}'.format(rr_id))

        (cd_id, tag) = rev_reg_id2cred_def_id_tag(rr_id)

        try:
            await self.get_cred_def(cd_id)
        except AbsentCredDef:
            LOGGER.debug(
                'Issuer._sync_revoc_for_issue <!< tails tree %s may be for another ledger; no cred def found on %s',
                self.dir_tails,
                cd_id)
            raise AbsentCredDef('Tails tree {} may be for another ledger; no cred def found on {}'.format(
                self.dir_tails,
                cd_id))

        with REVO_CACHE.lock:
            revo_cache_entry = REVO_CACHE.get(rr_id, None)
            tails = None if revo_cache_entry is None else revo_cache_entry.tails
            if tails is None:  #  it's a new revocation registry, or not yet set in cache
                try:
                    tails = await Tails(self.dir_tails, cd_id, tag).open()
                except AbsentTails:   # it's a new revocation registry
                    if self.rrbx:
                        await self._set_rev_reg(rr_id, rr_size)
                    else:
                        await self.rrb.create_rev_reg(rr_id, rr_size)
                        await self._send_rev_reg_def(rr_id)
                    tails = await Tails(self.dir_tails, cd_id, tag).open()  # symlink should exist now

                if revo_cache_entry is None:
                    REVO_CACHE[rr_id] = RevoCacheEntry(None, tails)
                else:
                    REVO_CACHE[rr_id].tails = tails

        LOGGER.debug('Issuer._sync_revoc_for_issue <<<')

    def path_tails(self, rr_id: str) -> str:
        """
        Return path to tails file for input revocation registry identifier.

        :param rr_id: revocation registry identifier of interest
        :return: path to tails file for input revocation registry identifier
        """

        LOGGER.debug('Issuer.path_tails >>>')

        if not ok_rev_reg_id(rr_id):
            LOGGER.debug('Issuer.path_tails <!< Bad rev reg id %s', rr_id)
            raise BadIdentifier('Bad rev reg id {}'.format(rr_id))

        rv = Tails.linked(self.dir_tails, rr_id)
        LOGGER.debug('Issuer.path_tails <<< %s', rv)
        return rv

    async def _create_cred_def(self, schema: dict, ledger_cred_def: dict, revo: bool) -> (str, bool):
        """
        Create credential definition in wallet as part of the send_cred_def() sequence.
        Return whether the private key for the cred def is OK to continue with the sequence,
        propagating the cred def and revocation registry info to the ledger.

        :param schema: schema on which to create cred def
        :param ledger_cred_def: credential definition as ledger has it (typically, None)
        :param revo: whether cred def supports revocation
        :return: cred def json and whether local cred def private key is OK, hence cred def is OK to send to the ledger
        """

        LOGGER.debug(
            'Issuer._create_cred_def >>> schema: %s, ledger_cred_def: %s, revo: %s',
            schema,
            ledger_cred_def,
            revo)

        cred_def_json = '{}'
        private_key_ok = True
        try:
            (_, cred_def_json) = await anoncreds.issuer_create_and_store_credential_def(
                self.wallet.handle,
                self.did,  # issuer DID
                json.dumps(schema),
                self.pool.protocol.cd_id_tag(False),  # expect only one cred def per schema and issuer
                'CL',
                json.dumps({'support_revocation': revo}))
            if ledger_cred_def:
                private_key_ok = False
                LOGGER.warning(
                    'New cred def on %s in wallet shadows existing one on ledger: private key not usable',
                    cred_def_id(self.did, schema['seqNo'], self.pool.protocol))
                    # carry on though, this anchor may have other capacities so public key may be good enough
        except IndyError as x_indy:
            if x_indy.error_code == ErrorCode.AnoncredsCredDefAlreadyExistsError:
                if ledger_cred_def:
                    LOGGER.info(
                        'Issuer wallet %s reusing existing cred def on schema %s version %s',
                        self.name,
                        schema['name'],
                        schema['version'])
                else:
                    LOGGER.debug('Issuer._create_cred_def <!< corrupt wallet %s', self.name)
                    raise CorruptWallet('Corrupt Issuer wallet {} has cred def on schema {} not on ledger'.format(
                        self.name,
                        schema['id']))
            else:
                LOGGER.debug(
                    'Issuer._create_cred_def <!< cannot store cred def in wallet %s: indy error code %s',
                    self.name,
                    x_indy.error_code)
                raise

        rv = (cred_def_json, private_key_ok)
        LOGGER.debug('Issuer._create_cred_def <<< %s', rv)
        return rv

    async def send_cred_def(self, s_id: str, revo: bool = True, rr_size: int = None) -> str:
        """
        Create a credential definition as Issuer, store it in its wallet, and send it to the ledger.

        Raise CorruptWallet for wallet not pertaining to current ledger, BadLedgerTxn on failure
        to send credential definition to ledger if need be, WalletState for closed wallet,
        or IndyError for any other failure to create and store credential definition in wallet.

        :param s_id: schema identifier
        :param revo: whether to support revocation for cred def
        :param rr_size: size of initial revocation registry (default as per RevRegBuilder.create_rev_reg()),
            if revocation supported
        :return: json credential definition as it appears on ledger
        """

        LOGGER.debug('Issuer.send_cred_def >>> s_id: %s, revo: %s, rr_size: %s', s_id, revo, rr_size)

        if not ok_schema_id(s_id):
            LOGGER.debug('Issuer.send_cred_def <!< Bad schema id %s', s_id)
            raise BadIdentifier('Bad schema id {}'.format(s_id))

        if not self.wallet.handle:
            LOGGER.debug('Issuer.send_cred_def <!< Wallet %s is closed', self.name)
            raise WalletState('Wallet {} is closed'.format(self.name))

        if not self.pool:
            LOGGER.debug('Issuer.send_cred_def <!< issuer %s has no pool', self.name)
            raise AbsentPool('Issuer {} has no pool: cannot send cred def'.format(self.name))

        rv_json = json.dumps({})
        schema_json = await self.get_schema(schema_key(s_id))
        schema = json.loads(schema_json)

        cd_id = cred_def_id(self.did, schema['seqNo'], self.pool.protocol)
        private_key_ok = True
        with CRED_DEF_CACHE.lock:
            try:
                rv_json = await self.get_cred_def(cd_id)
                LOGGER.info(
                    'Cred def on schema %s version %s already exists on ledger; Issuer %s not sending another',
                    schema['name'],
                    schema['version'],
                    self.name)
            except AbsentCredDef:
                pass  # OK - about to create, store, and send it

            (cred_def_json, private_key_ok) = await self._create_cred_def(schema, json.loads(rv_json), revo)

            if not json.loads(rv_json):  # checking the ledger returned no cred def: send it
                req_json = await ledger.build_cred_def_request(self.did, cred_def_json)
                await self._sign_submit(req_json)

                for _ in range(16):  # reasonable timeout
                    try:
                        rv_json = await self.get_cred_def(cd_id)  # adds to cache
                        break
                    except AbsentCredDef:
                        await asyncio.sleep(1)
                        LOGGER.info('Sent cred def %s to ledger, waiting 1s for its appearance', cd_id)

                if not rv_json:
                    LOGGER.debug('Issuer.send_cred_def <!< timed out waiting on sent cred_def %s', cd_id)
                    raise BadLedgerTxn('Timed out waiting on sent cred_def {}'.format(cd_id))

                if revo:  # create new rev reg for tag '0'
                    if self.rrbx:
                        (_, rr_size_suggested) = Tails.next_tag(self.dir_tails, cd_id)
                        self.rrb.mark_in_progress(rev_reg_id(cd_id, '0'), rr_size or rr_size_suggested)

                    await self._sync_revoc_for_issue(rev_reg_id(cd_id, '0'), rr_size)  # sync rev reg on tag '0'

        if revo and private_key_ok:
            for tag in [str(t) for t in range(1, int(Tails.next_tag(self.dir_tails, cd_id)[0]))]:  # '1' to next-1
                await self._sync_revoc_for_issue(rev_reg_id(cd_id, tag), rr_size if tag == '0' else None)

        makedirs(join(self.dir_tails, cd_id), exist_ok=True)  # dir required for box id collection, revo or not

        LOGGER.debug('Issuer.send_cred_def <<< %s', rv_json)
        return rv_json

    async def create_cred_offer(self, schema_seq_no: int) -> str:
        """
        Create credential offer as Issuer for given schema.

        Raise CorruptWallet if the wallet has no private key for the corresponding credential definition.
        Raise WalletState for closed wallet.

        :param schema_seq_no: schema sequence number
        :return: credential offer json for use in storing credentials at HolderProver.
        """

        LOGGER.debug('Issuer.create_cred_offer >>> schema_seq_no: %s', schema_seq_no)

        if not self.wallet.handle:
            LOGGER.debug('Issuer.create_cred_offer <!< Wallet %s is closed', self.name)
            raise WalletState('Wallet {} is closed'.format(self.name))

        if not self.pool:
            LOGGER.debug('Issuer.create_cred_offer <!< issuer %s has no pool', self.name)
            raise AbsentPool('Issuer {} has no pool: cannot create cred offer'.format(self.name))

        rv = None
        cd_id = cred_def_id(self.did, schema_seq_no, self.pool.protocol)
        try:
            rv = await anoncreds.issuer_create_credential_offer(self.wallet.handle, cd_id)
        except IndyError as x_indy:
            if x_indy.error_code == ErrorCode.WalletItemNotFound:
                LOGGER.debug(
                    'Issuer.create_cred_offer <!< did not issue cred definition from wallet %s',
                    self.name)
                raise CorruptWallet('Cannot create cred offer: did not issue cred definition from wallet {}'.format(
                    self.name))
            LOGGER.debug(
                'Issuer.create_cred_offer <!< cannot create cred offer, indy error code %s',
                x_indy.error_code)
            raise

        LOGGER.debug('Issuer.create_cred_offer <<< %s', rv)
        return rv

    async def create_cred(
            self,
            cred_offer_json,
            cred_req_json: str,
            cred_attrs: dict,
            rr_size: int = None) -> (str, str):
        """
        Create credential as Issuer out of credential request and dict of key:value (raw, unencoded)
        entries for attributes.

        Return credential json, and if cred def supports revocation, credential revocation identifier.
        Raise WalletState for closed wallet.

        If the credential definition supports revocation, and the current revocation registry is full,
        the processing creates a new revocation registry en passant. Depending on the revocation
        registry size (by default starting at 64 and doubling iteratively through a maximum of 100000)
        and the revocation registry builder posture (see RevRegBuilder.__init__()), this operation may
        delay credential creation by several seconds. The use of an external revocation registry builder
        runs a parallel process, skirting this delay, but is more costly at initialization.

        :param cred_offer_json: credential offer json as created by Issuer
        :param cred_req_json: credential request json as created by HolderProver
        :param cred_attrs: dict mapping each attribute to its original value (the operation encodes it); e.g.,

        ::

            {
                'favourite_drink': 'martini',
                'height': 180,
                'last_visit_date': '2017-12-31',
                'weaknesses': None
            }

        :param rr_size: size of new revocation registry (default as per RevRegBuilder.create_rev_reg()) if necessary
        :return: tuple with newly issued credential json, credential revocation identifier (if cred def
            supports revocation, None otherwise).
        """

        LOGGER.debug(
            'Issuer.create_cred >>> cred_offer_json: %s, cred_req_json: %s, cred_attrs: %s, rr_size: %s',
            cred_offer_json,
            cred_req_json,
            cred_attrs,
            rr_size)

        if not self.wallet.handle:
            LOGGER.debug('Issuer.create_cred <!< Wallet %s is closed', self.name)
            raise WalletState('Wallet {} is closed'.format(self.name))

        cd_id = json.loads(cred_offer_json)['cred_def_id']
        if not ok_cred_def_id(cd_id):
            LOGGER.debug('Issuer.create_cred <!< Bad cred def id %s', cd_id)
            raise BadIdentifier('Bad cred def id {}'.format(cd_id))

        cred_def = json.loads(await self.get_cred_def(cd_id))  # ensure cred def is in cache

        if 'revocation' in cred_def['value']:
            with REVO_CACHE.lock:
                rr_id = Tails.current_rev_reg_id(self.dir_tails, cd_id)
                tails = REVO_CACHE[rr_id].tails
                assert tails  # at (re)start, at cred def, Issuer sync_revoc_for_issue() sets this index in revo cache

                try:
                    (cred_json, cred_revoc_id, _) = await anoncreds.issuer_create_credential(  # issue by default to rr
                        self.wallet.handle,
                        cred_offer_json,
                        cred_req_json,
                        json.dumps({k: cred_attr_value(cred_attrs[k]) for k in cred_attrs}),
                        rr_id,
                        tails.reader_handle)
                    rv = (cred_json, cred_revoc_id)

                except IndyError as x_indy:
                    if x_indy.error_code == ErrorCode.AnoncredsRevocationRegistryFullError:
                        (tag, rr_size_suggested) = Tails.next_tag(self.dir_tails, cd_id)
                        rr_id = rev_reg_id(cd_id, tag)
                        if self.rrbx:
                            await self._set_rev_reg(rr_id, rr_size)
                        else:
                            await self.rrb.create_rev_reg(rr_id, rr_size or rr_size_suggested)
                            await self._send_rev_reg_def(rr_id)

                        REVO_CACHE[rr_id].tails = await Tails(self.dir_tails, cd_id).open()  # symlink OK now
                        return await self.create_cred(cred_offer_json, cred_req_json, cred_attrs)

                    LOGGER.debug('Issuer.create_cred <!< cannot create cred, indy error code %s', x_indy.error_code)
                    raise
        else:
            try:
                (cred_json, _, _) = await anoncreds.issuer_create_credential(
                    self.wallet.handle,
                    cred_offer_json,
                    cred_req_json,
                    json.dumps({k: cred_attr_value(cred_attrs[k]) for k in cred_attrs}),
                    None,
                    None)
                rv = (cred_json, None)
            except IndyError as x_indy:
                LOGGER.debug('Issuer.create_cred <!< cannot create cred, indy error code %s', x_indy.error_code)
                raise

        LOGGER.debug('Issuer.create_cred <<< %s', rv)
        return rv

    async def revoke_cred(self, rr_id: str, cr_id) -> int:
        """
        Revoke credential that input revocation registry identifier and
        credential revocation identifier specify.

        Return (epoch seconds) time of revocation.

        Raise AbsentTails if no tails file is available for input revocation registry identifier.
        Raise WalletState for closed wallet.
        Raise BadRevocation if issuer cannot revoke specified credential for any other reason
        (e.g., did not issue it, already revoked it).

        :param rr_id: revocation registry identifier
        :param cr_id: credential revocation identifier
        :return: time of revocation, in epoch seconds
        """

        LOGGER.debug('Issuer.revoke_cred >>> rr_id: %s, cr_id: %s', rr_id, cr_id)

        if not self.wallet.handle:
            LOGGER.debug('Issuer.revoke_cred <!< Wallet %s is closed', self.name)
            raise WalletState('Wallet {} is closed'.format(self.name))

        if not ok_rev_reg_id(rr_id):
            LOGGER.debug('Issuer.revoke_cred <!< Bad rev reg id %s', rr_id)
            raise BadIdentifier('Bad rev reg id {}'.format(rr_id))

        tails_reader_handle = (await Tails(
            self.dir_tails,
            *rev_reg_id2cred_def_id_tag(rr_id)).open()).reader_handle
        try:
            rrdelta_json = await anoncreds.issuer_revoke_credential(
                self.wallet.handle,
                tails_reader_handle,
                rr_id,
                cr_id)
        except IndyError as x_indy:
            LOGGER.debug(
                'Issuer.revoke_cred <!< Could not revoke revoc reg id %s, cred rev id %s: indy error code %s',
                rr_id,
                cr_id,
                x_indy.error_code)
            raise BadRevocation(
                'Could not revoke revoc reg id {}, cred rev id {}: indy error code {}'.format(
                    rr_id,
                    cr_id,
                    x_indy.error_code))

        rr_ent_req_json = await ledger.build_revoc_reg_entry_request(self.did, rr_id, 'CL_ACCUM', rrdelta_json)
        resp_json = await self._sign_submit(rr_ent_req_json)  # raises AbsentPool or ClosedPool if applicable
        resp = json.loads(resp_json)

        rv = self.pool.protocol.txn2epoch(resp)
        LOGGER.debug('Issuer.revoke_cred <<< %s', rv)
        return rv

    async def get_box_ids_issued(self) -> str:
        """
        Return json object on lists of all unique box identifiers (schema identifiers,
        credential definition identifiers, and revocation registry identifiers) for
        all credential definitions and credentials issued; e.g.,

        ::

            {
                "schema_id": [
                    "R17v42T4pk...:2:tombstone:1.2",
                    ...
                ],
                "cred_def_id": [
                    "R17v42T4pk...:3:CL:19:tag",
                    ...
                ]
                "rev_reg_id": [
                    "R17v42T4pk...:4:R17v42T4pk...:3:CL:19:tag:CL_ACCUM:0",
                    "R17v42T4pk...:4:R17v42T4pk...:3:CL:19:tag:CL_ACCUM:1",
                    ...
                ]
            }

        An issuer must issue a credential definition to include its schema identifier
        in the returned values; the schema identifier in isolation belongs properly
        to an Origin, not necessarily to an Issuer.

        The operation may be useful for a Verifier anchor going off-line to seed its
        cache before doing so.

        :return: tuple of sets for schema ids, cred def ids, rev reg ids
        """

        LOGGER.debug('Issuer.get_box_ids_issued >>>')

        cd_ids = [
            d for d in listdir(self.dir_tails) if isdir(join(self.dir_tails, d)) and ok_cred_def_id(d, self.did)]
        s_ids = []
        for cd_id in cd_ids:
            try:
                s_ids.append(json.loads(await self.get_schema(cred_def_id2seq_no(cd_id)))['id'])
            except AbsentSchema:
                LOGGER.error(
                    'Issuer %s has issued cred def %s but no corresponding schema on ledger',
                    self.name,
                    cd_id)
        rr_ids = [basename(link) for link in Tails.links(self.dir_tails, self.did)]

        rv = json.dumps({
            'schema_id': s_ids,
            'cred_def_id': cd_ids,
            'rev_reg_id': rr_ids
        })
        LOGGER.debug('Issuer.get_box_ids_issued <<< %s', rv)
        return rv



"""
Copyright 2017-2019 Government of Canada - Public Services and Procurement Canada - buyandsell.gc.ca

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""


import json

from collections import namedtuple
from enum import Enum

from von_anchor.indytween import SchemaKey


ProtocolMap = namedtuple('ProtocolMap', 'name indy')


class Protocol(Enum):
    """
    Class encapsulating indy-node transaction particulars by protocol version.
    """

    V_13 = ProtocolMap('1.3', 1)
    V_14 = ProtocolMap('1.4', 2)
    V_15 = ProtocolMap('1.5', 2)
    V_16 = ProtocolMap('1.6', 2)
    V_17 = ProtocolMap('1.7', 2)
    V_18 = ProtocolMap('1.8', 2)
    DEFAULT = ProtocolMap('1.8', 2)

    @staticmethod
    def get(version: str) -> 'Protocol':
        """
        Return enum instance corresponding to input version value ('1.6' etc.)
        """

        return Protocol.V_13 if version == Protocol.V_13.value.name else Protocol.DEFAULT

    def __str__(self) -> str:
        return self.name

    def indy(self) -> int:
        """
        Return indy integer mapping for protocol.

        :return: indy integer mapping for protocol
        """

        return self.value.indy

    def cd_id_tag(self, for_box_id: bool = False) -> str:
        """
        Return (place-holder) credential definition identifier tag for current version of node protocol.
        At present, von_anchor always uses the tag of 'tag' if the protocol calls for one.

        :param for_box_id: whether to prefix a colon, if current protocol uses one, in constructing
            a cred def id or rev reg id.
        :return: cred def id tag
        """

        if for_box_id:
            return '' if self == Protocol.V_13 else ':tag'
        return 'tag'

    def cred_def_id(self, issuer_did: str, schema_seq_no: int) -> str:
        """
        Return credential definition identifier for input issuer DID and schema sequence number.

        :param issuer_did: DID of credential definition issuer
        :param schema_seq_no: schema sequence number
        :return: credential definition identifier
        """

        return '{}:3:CL:{}{}'.format(  # 3 marks indy cred def id, CL is sig type
            issuer_did,
            schema_seq_no,
            self.cd_id_tag(True))

    def txn_data2schema_key(self, txn: dict) -> SchemaKey:
        """
        Return schema key from ledger transaction data.

        :param txn: get-schema transaction (by sequence number)
        :return: schema key identified
        """

        rv = None
        if self == Protocol.V_13:
            rv = SchemaKey(txn['identifier'], txn['data']['name'], txn['data']['version'])
        else:
            txn_txn = txn.get('txn', None) or txn  # may have already run this txn through txn2data() below
            rv = SchemaKey(
                txn_txn['metadata']['from'],
                txn_txn['data']['data']['name'],
                txn_txn['data']['data']['version'])

        return rv

    def txn2data(self, txn: dict) -> str:
        """
        Given ledger transaction, return its data json.

        :param txn: transaction as dict
        :return: transaction data json
        """

        rv_json = json.dumps({})
        if self == Protocol.V_13:
            rv_json = json.dumps(txn['result'].get('data', {}))
        else:
            rv_json = json.dumps((txn['result'].get('data', {}) or {}).get('txn', {}))  # "data": null for no such txn

        return rv_json

    def txn2epoch(self, txn: dict) -> int:
        """
        Given ledger transaction, return its epoch time.

        :param txn: transaction as dict
        :return: transaction time
        """

        rv = None
        if self == Protocol.V_13:
            rv = txn['result']['txnTime']
        else:
            rv = txn['result']['txnMetadata']['txnTime']

        return rv



"""
Copyright 2017-2019 Government of Canada - Public Services and Procurement Canada - buyandsell.gc.ca

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""


import json
import re

from copy import deepcopy
from typing import Sequence, Union

from base58 import alphabet, b58decode

from von_anchor.error import BadIdentifier
from von_anchor.nodepool import Protocol
from von_anchor.indytween import Role, Predicate, SchemaKey


B58 = alphabet if isinstance(alphabet, str) else alphabet.decode('ascii')


def schema_id(origin_did: str, name: str, version: str) -> str:
    """
    Return schema identifier for input origin DID, schema name, and schema version.

    :param origin_did: DID of schema originator
    :param name: schema name
    :param version: schema version
    :return: schema identifier
    """

    return '{}:2:{}:{}'.format(origin_did, name, version)  # 2 marks indy-sdk schema id


def ok_wallet_reft(token: str) -> bool:
    """
    Whether input token looks like a valid wallet credential identifier
    (aka wallet referent, wallet cred id, wallet cred uuid).

    :param token: candidate string
    :return: whether input token looks like a valid wallet credential identifier
    """

    return bool(re.match(r'[0-9a-f]{8}-([0-9a-f]{4}-){3}[0-9a-f]{12}$', token or ''))


def ok_role(token: str) -> bool:
    """
    Whether input role is valid to mark indy anchor role on ledger.

    :param token: candidate string
    :return: whether input token looks like a valid indy anchor role
    """

    return Role.get(token) is not None


def ok_endpoint(token: str) -> bool:
    """
    Whether input token looks like a valid indy endpoint (<ip-address>:<port>).

    :param token: candidate string
    :return: whether input token looks like a valid indy endpoint
    """

    return bool(re.match(
        r'((25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.){3}(25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?):[0-9]+$',
        token or ''))


def ok_did(token: str) -> bool:
    """
    Whether input token looks like a valid distributed identifier.

    :param token: candidate string
    :return: whether input token looks like a valid schema identifier
    """

    try:
        return len(b58decode(token)) == 16 if token else False
    except ValueError:
        return False


def ok_schema_id(token: str) -> bool:
    """
    Whether input token looks like a valid schema identifier;
    i.e., <issuer-did>:2:<name>:<version>.

    :param token: candidate string
    :return: whether input token looks like a valid schema identifier
    """

    return bool(re.match('[{}]{{21,22}}:2:.+:[0-9.]+$'.format(B58), token or ''))


def schema_key(s_id: str) -> SchemaKey:
    """
    Return schema key (namedtuple) convenience for schema identifier components.

    :param s_id: schema identifier
    :return: schema key (namedtuple) object
    """

    s_key = s_id.split(':')
    s_key.pop(1)  # take out indy-sdk schema marker: 2 marks indy-sdk schema id

    return SchemaKey(*s_key)


def cred_def_id(issuer_did: str, schema_seq_no: int, protocol: Protocol = None) -> str:
    """
    Return credential definition identifier for input issuer DID and schema sequence number.

    Implementation passes to NodePool Protocol.

    :param issuer_did: DID of credential definition issuer
    :param schema_seq_no: schema sequence number
    :param protocol: indy protocol version
    :return: credential definition identifier
    """

    return (protocol or Protocol.DEFAULT).cred_def_id(issuer_did, schema_seq_no)


def ok_cred_def_id(token: str, issuer_did: str = None) -> bool:
    """
    Whether input token looks like a valid credential definition identifier from input issuer DID (default any); i.e.,
    <issuer-did>:3:CL:<schema-seq-no>:<cred-def-id-tag> for protocol >= 1.4, or
    <issuer-did>:3:CL:<schema-seq-no> for protocol == 1.3.

    :param token: candidate string
    :param issuer_did: issuer DID to match, if specified
    :return: whether input token looks like a valid credential definition identifier
    """

    cd_id_m = re.match('([{}]{{21,22}}):3:CL:[1-9][0-9]*(:.+)?$'.format(B58), token or '')
    return bool(cd_id_m) and ((not issuer_did) or cd_id_m.group(1) == issuer_did)


def cred_def_id2seq_no(cd_id: str) -> int:
    """
    Given a credential definition identifier, return its schema sequence number.
    Raise BadIdentifier on input that is not a credential definition identifier.

    :param cd_id: credential definition identifier
    :return: sequence number
    """

    if ok_cred_def_id(cd_id):
        return int(cd_id.split(':')[3])  # sequence number is token at 0-based position 3
    raise BadIdentifier('Bad credential definition identifier {}'.format(cd_id))


def rev_reg_id(cd_id: str, tag: Union[str, int]) -> str:
    """
    Given a credential definition identifier and a tag, return the corresponding
    revocation registry identifier, repeating the issuer DID from the
    input identifier.

    :param cd_id: credential definition identifier
    :param tag: tag to use
    :return: revocation registry identifier
    """

    return '{}:4:{}:CL_ACCUM:{}'.format(cd_id.split(":", 1)[0], cd_id, tag)  # 4 marks rev reg def id


def ok_rev_reg_id(token: str, issuer_did: str = None) -> bool:
    """
    Whether input token looks like a valid revocation registry identifier from input issuer DID (default any); i.e.,
    <issuer-did>:4:<issuer-did>:3:CL:<schema-seq-no>:<cred-def-id-tag>:CL_ACCUM:<rev-reg-id-tag> for protocol >= 1.4, or
    <issuer-did>:4:<issuer-did>:3:CL:<schema-seq-no>:CL_ACCUM:<rev-reg-id-tag> for protocol == 1.3.

    :param token: candidate string
    :param issuer_did: issuer DID to match, if specified
    :return: whether input token looks like a valid revocation registry identifier
    """

    rr_id_m = re.match(
        '([{0}]{{21,22}}):4:([{0}]{{21,22}}):3:CL:[1-9][0-9]*(:.+)?:CL_ACCUM:.+$'.format(B58),
        token or '')
    return bool(rr_id_m) and ((not issuer_did) or (rr_id_m.group(1) == issuer_did and rr_id_m.group(2) == issuer_did))


def rev_reg_id2cred_def_id(rr_id: str) -> str:
    """
    Given a revocation registry identifier, return its corresponding credential definition identifier.
    Raise BadIdentifier if input is not a revocation registry identifier.

    :param rr_id: revocation registry identifier
    :return: credential definition identifier
    """

    if ok_rev_reg_id(rr_id):
        return ':'.join(rr_id.split(':')[2:-2])  # rev reg id comprises (prefixes):<cred_def_id>:(suffixes)
    raise BadIdentifier('Bad revocation registry identifier {}'.format(rr_id))


def rev_reg_id2tag(rr_id: str) -> str:
    """
    Given a revocation registry identifier, return its corresponding (stringified int) tag.

    :param rr_id: revocation registry identifier
    :return: tag
    """

    return str(rr_id.split(':')[-1])  # tag is last token


def rev_reg_id2cred_def_id_tag(rr_id: str) -> (str, str):
    """
    Given a revocation registry identifier, return its corresponding credential definition identifier and
    (stringified int) tag. Raise BadIdentifier if input is not a revocation registry identifier.

    :param rr_id: revocation registry identifier
    :return: credential definition identifier and tag
    """

    if ok_rev_reg_id(rr_id):
        return (
            ':'.join(rr_id.split(':')[2:-2]),  # rev reg id comprises (prefixes):<cred_def_id>:(suffixes)
            str(rr_id.split(':')[-1])  # tag is last token
        )
    raise BadIdentifier('Bad revocation registry identifier {}'.format(rr_id))


def iter_briefs(briefs: Union[dict, Sequence[dict]]) -> tuple:
    """
    Given a cred-brief/cred-info, an sequence thereof, or cred-brief-dict
    (as HolderProver.get_cred_briefs_by_proof_req_q() returns), return tuple with
    all contained cred-briefs.

    :param briefs: cred-brief/cred-info, sequence thereof, or cred-brief-dict
    :return: tuple of cred-briefs
    """

    if isinstance(briefs, dict):
        if all(ok_wallet_reft(k) for k in briefs):
            return tuple(briefs.values())
        return (briefs,)
    return tuple(briefs)


def box_ids(briefs: Union[dict, Sequence[dict]], cred_ids: Union[Sequence[str], str] = None) -> dict:
    """
    Given one or more cred-briefs/cred-infos, and an optional sequence of credential identifiers
    (aka wallet cred ids, referents; specify None to include all), return dict mapping each
    credential identifier to a box ids structure (i.e., a dict specifying its corresponding
    schema identifier, credential definition identifier, and revocation registry identifier,
    the latter being None if cred def does not support revocation).

    :param briefs: cred-brief/cred-info, sequence thereof, or cred-brief-dict
    :param cred_ids: credential identifier or sequence thereof for which to find corresponding
        schema identifiers, None for all
    :return: dict mapping each credential identifier to its corresponding box ids (empty dict if
        no matching credential identifiers present)
    """

    rv = {}
    for brief in iter_briefs(briefs):
        cred_info = brief.get('cred_info', {}) or brief  # briefs could be cred-infos or cred-briefs
        cred_id = cred_info['referent']
        if ((cred_id not in rv) and (not cred_ids or cred_id in [cred_ids, [cred_ids]][isinstance(cred_ids, str)])):
            rv[cred_id] = {
                'schema_id': cred_info['schema_id'],
                'cred_def_id': cred_info['cred_def_id'],
                'rev_reg_id': cred_info['rev_reg_id']
            }

    return rv


def prune_creds_json(creds: dict, cred_ids: set) -> str:
    """
    Strip all creds out of the input json structure that do not match any of the input credential identifiers.

    :param creds: indy-sdk creds structure
    :param cred_ids: the set of credential identifiers of interest
    :return: the reduced creds json
    """

    rv = deepcopy(creds)
    for key in ('attrs', 'predicates'):
        for attr_uuid, creds_by_uuid in rv[key].items():
            rv[key][attr_uuid] = [cred for cred in creds_by_uuid if cred['cred_info']['referent'] in cred_ids]

        empties = [attr_uuid for attr_uuid in rv[key] if not rv[key][attr_uuid]]
        for attr_uuid in empties:
            del rv[key][attr_uuid]

    return json.dumps(rv)


def proof_req_infos2briefs(proof_req: dict, infos: Union[dict, Sequence[dict]]) -> list:
    """
    Given a proof request and corresponding cred-info(s), return a list of cred-briefs
    (i.e., cred-info plus interval).

    The proof request must have cred def id restrictions on all requested attribute specifications.

    :param proof_req: proof request
    :param infos: cred-info or sequence thereof; e.g.,

    ::

        [
            {
                'attrs': {
                    'auditDate': '2018-07-30',
                    'greenLevel': 'Silver',
                    'legalName': 'Tart City'
                },
                'cred_rev_id': '48',
                'cred_def_id': 'WgWxqztrNooG92RXvxSTWv:3:CL:17:tag',
                'referent': 'c15674a9-7321-440d-bbed-e1ac9273abd5',
                'rev_reg_id': 'WgWxqztrNooG92RXvxSTWv:4:WgWxqztrNooG92RXvxSTWv:3:CL:17:tag:CL_ACCUM:0',
                'schema_id': 'WgWxqztrNooG92RXvxSTWv:2:green:1.0'
            },
            ...
        ]

    :return: list of cred-briefs
    """

    rv = []
    refts = proof_req_attr_referents(proof_req)
    for info in iter_briefs(infos):
        if info['cred_def_id'] not in refts:
            continue
        brief = {
            'cred_info': info,
            'interval': {}
        }
        fro = None
        to = None
        for uuid in refts[info['cred_def_id']].values():
            interval = proof_req['requested_attributes'][uuid].get('non_revoked', {})
            if 'from' in interval:
                fro = min(fro or interval['from'], interval['from'])
            if 'to' in interval:
                to = max(to or interval['to'], interval['to'])

        if to:
            brief['interval']['to'] = to
        if fro:
            brief['interval']['from'] = fro
        if not brief['interval']:
            brief['interval'] = None

        rv.append(brief)

    return rv

def proof_req_briefs2req_creds(proof_req: dict, briefs: Union[dict, Sequence[dict]]) -> dict:
    """
    Given a proof request and cred-brief(s), return a requested-creds structure.

    The proof request must have cred def id restrictions on all requested attribute specifications.

    :param proof_req: proof request
    :param briefs: credential brief, sequence thereof (as indy-sdk wallet credential search returns),
        or cred-brief-dict (as HolderProver.get_cred_briefs_for_proof_req_q() returns); e.g.,

    ::

        [
            {
                "cred_info": {
                    "cred_rev_id": "149",
                    "cred_def_id": "LjgpST2rjsoxYegQDRm7EL:3:CL:15:tag",
                    "schema_id": "LjgpST2rjsoxYegQDRm7EL:2:bc-reg:1.0",
                    "rev_reg_id": "LjgpST2rjsoxYegQDRm7EL:4:LjgpST2rjsoxYegQDRm7EL:3:CL:15:tag:CL_ACCUM:1",
                    "referent": "43f8dc18-ac00-4b72-8a96-56f47dba77ca",
                    "attrs": {
                        "busId": "11144444",
                        "endDate": "",
                        "id": "3",
                        "effectiveDate": "2012-12-01",
                        "jurisdictionId": "1",
                        "orgTypeId": "2",
                        "legalName": "Tart City"
                    }
                },
                "interval": {
                    "to": 1532448939,
                    "from": 1234567890
                }
            },
            ...
        ]

    :return: indy-sdk requested creds json to pass to proof creation request; e.g.,

    ::

        {
            "requested_attributes": {
                "15_endDate_uuid": {
                    "timestamp": 1532448939,
                    "cred_id": "43f8dc18-ac00-4b72-8a96-56f47dba77ca",
                    "revealed": true
                },
                "15_id_uuid": {
                    "timestamp": 1532448939,
                    "cred_id": "43f8dc18-ac00-4b72-8a96-56f47dba77ca",
                    "revealed": true
                },
                "15_effectiveDate_uuid": {
                    "timestamp": 1532448939,
                    "cred_id": "43f8dc18-ac00-4b72-8a96-56f47dba77ca",
                    "revealed": true
                },
                "15_busId_uuid": {
                    "timestamp": 1532448939,
                    "cred_id": "43f8dc18-ac00-4b72-8a96-56f47dba77ca",
                    "revealed": true
                },
                "15_orgTypeId_uuid": {
                    "timestamp": 1532448939,
                    "cred_id": "43f8dc18-ac00-4b72-8a96-56f47dba77ca",
                    "revealed": false
                },
                "15_jurisdictionId_uuid": {
                    "timestamp": 1532448939,
                    "cred_id": "43f8dc18-ac00-4b72-8a96-56f47dba77ca",
                    "revealed": true
                },
                "15_legalName_uuid": {
                    "timestamp": 1532448939,
                    "cred_id": "43f8dc18-ac00-4b72-8a96-56f47dba77ca",
                    "revealed": true
                }
            },
            "requested_predicates": {
                "15_orgTypeId_GE_uuid": {
                    "timestamp": 1532448939,
                    "cred_id": "43f8dc18-ac00-4b72-8a96-56f47dba77ca",
                }
            },
            "self_attested_attributes": {}
        }

    """

    rv = {
        'self_attested_attributes': {},
        'requested_attributes': {},
        'requested_predicates': {}
    }

    attr_refts = proof_req_attr_referents(proof_req)
    pred_refts = proof_req_pred_referents(proof_req)
    for brief in iter_briefs(briefs):
        cred_info = brief['cred_info']
        timestamp = (brief['interval'] or {}).get('to', None)
        for attr in cred_info['attrs']:
            if attr in attr_refts.get(cred_info['cred_def_id'], {}):
                req_attr = {
                    'cred_id': cred_info['referent'],
                    'revealed': attr not in pred_refts.get(cred_info['cred_def_id'], {}),
                    'timestamp': timestamp
                }
                if not timestamp:
                    req_attr.pop('timestamp')
                rv['requested_attributes'][attr_refts[cred_info['cred_def_id']][attr]] = req_attr
            if attr in pred_refts.get(cred_info['cred_def_id'], {}):
                for uuid in pred_refts[cred_info['cred_def_id']][attr]:
                    req_pred = {
                        'cred_id': cred_info['referent'],
                        'timestamp': timestamp
                    }
                    if not timestamp:
                        req_pred.pop('timestamp')
                    rv['requested_predicates'][uuid] = req_pred
    return rv


def creds_display(creds: Union[dict, Sequence[dict]], filt: dict = None, filt_dflt_incl: bool = False) -> dict:
    """
    Find indy-sdk creds matching input filter from within input creds structure,
    sequence of cred-briefs/cred-infos, or cred-brief-dict.  Return human-legible summary.

    :param creds: creds structure, cred-brief/cred-info or sequence thereof,
        or cred-brief-dict; e.g., creds

    ::

        {
            "attrs": {
                "attr0_uuid": [
                    {
                        "interval": null,
                        "cred_info": {
                            "attrs": {
                                "attr0": "2",
                                "attr1": "Hello",
                                "attr2": "World"
                            },
                            "referent": "00000000-0000-0000-0000-000000000000",
                            "schema_id": "Q4zqM7aXqm7gDQkUVLng9h:2:bc-reg:1.0",
                            "cred_def_id": "Q4zqM7aXqm7gDQkUVLng9h:3:CL:18:tag",
                            "cred_rev_id": null,
                            "rev_reg_id": null
                        }
                    },
                    {
                        "interval": null,
                        "cred_info": {
                            "attrs": {
                                "attr0": "1",
                                "attr1": "Nice",
                                "attr2": "Tractor"
                            },
                            "referent": "00000000-0000-0000-0000-111111111111",
                            "schema_id": "Q4zqM7aXqm7gDQkUVLng9h:2:bc-reg:1.0",
                            "cred_def_id": "Q4zqM7aXqm7gDQkUVLng9h:3:CL:18:tag",
                            "cred_rev_id": null,
                            "rev_reg_id": null
                        }
                    }
                ],
                "attr1_uuid": [
                    {
                        "interval": null,
                        "cred_info": {
                            "attrs": {
                                "attr0": "2",
                                "attr1": "Hello",
                                "attr2": "World"
                            },
                            "referent": "00000000-0000-0000-0000-000000000000",
                            "schema_id": "Q4zqM7aXqm7gDQkUVLng9h:2:bc-reg:1.0",
                            "cred_def_id": "Q4zqM7aXqm7gDQkUVLng9h:3:CL:18:tag",
                            "cred_rev_id": null,
                            "rev_reg_id": null
                        }
                    },
                    {
                        "interval": null,
                        "cred_info": {
                            "attrs": {
                                "attr0": "1",
                                "attr1": "Nice",
                                "attr2": "Tractor"
                            },
                            "referent": "00000000-0000-0000-0000-111111111111",
                            "schema_id": "Q4zqM7aXqm7gDQkUVLng9h:2:bc-reg:1.0",
                            "cred_def_id": "Q4zqM7aXqm7gDQkUVLng9h:3:CL:18:tag",
                            "cred_rev_id": null,
                            "rev_reg_id": null
                        }
                    }
                ],
                "attr2_uuid": [
                    ...
                ]
            }
        }

    :param filt: filter for matching attributes and values; dict (None or empty for no filter, matching all)
        mapping each cred def identifier to dict mapping attributes to values to match; e.g.,

    ::

        {
            'Q4zqM7aXqm7gDQkUVLng9h:3:CL:18:tag': {
                'attr0': 1,  # operation stringifies en passant
                'attr1': 'Nice'
            },
            ...
        }

    :param filt_dflt_incl: whether to include (True) all attributes for schema that filter does not identify
        or to exclude (False) all such attributes
    :return: human-legible dict mapping credential identifiers to human-readable credential briefs
        (not proper indy-sdk creds structures) for creds matching input filter
    """

    def _add(briefs):
        nonlocal rv, filt
        for brief in briefs:
            cred_info = brief.get('cred_info', {}) or brief  # briefs could be cred-infos or cred-briefs
            if cred_info['referent'] in rv:
                continue
            cred_cd_id = cred_info['cred_def_id']
            if (not filt) or (filt_dflt_incl and cred_cd_id not in filt):
                rv[cred_info['referent']] = cred_info
                continue
            if filt and cred_cd_id in filt:
                if ({k: str(filt[cred_cd_id][k]) for k in filt[cred_cd_id]}.items() <= cred_info['attrs'].items()):
                    rv[cred_info['referent']] = cred_info

    rv = {}
    if filt is None:
        filt = {}
    if isinstance(creds, dict):
        if all(ok_wallet_reft(k) for k in creds):
            _add(creds.values())
        else:
            for uuid2briefs in (creds.get('attrs', {}), creds.get('predicates', {})):
                for briefs in uuid2briefs.values():
                    _add(briefs)
    else:
        _add(creds)
    return rv


def proof_req2wql_all(proof_req: dict, x_cd_ids: Union[str, Sequence[str]] = None) -> dict:
    """
    Given a proof request and a list of cred def ids to omit, return an extra WQL query dict
    that will find all corresponding credentials in search.

    The proof request must have cred def id restrictions on all requested attribute specifications.
    At present, the utility does not support predicates.

    :param proof_req: proof request
    :param x_cd_ids: cred def identifier or sequence thereof to omit
    :return: extra WQL dict to fetch all corresponding credentials in search.
    """

    rv = {}
    attr_refts = proof_req_attr_referents(proof_req)
    for cd_id in [k for k in attr_refts if k not in ([x_cd_ids] if isinstance(x_cd_ids, str) else x_cd_ids or [])]:
        rv[set(attr_refts[cd_id].values()).pop()] = {"cred_def_id": cd_id}

    return rv


def proof_req_attr_referents(proof_req: dict) -> dict:
    """
    Given a proof request with all requested attributes having cred def id restrictions,
    return its attribute referents by cred def id and attribute.

    The returned structure can be useful in populating the extra WQL query parameter
    in the credential search API.

    :param proof_req: proof request with all requested attribute specifications having cred def id restriction; e.g.,

    ::

        {
            'name": 'proof_req',
            'version': '0.0',
            'requested_attributes': {
                '18_greenLevel_uuid': {
                    'restrictions': [
                        {
                            'cred_def_id': 'WgWxqztrNooG92RXvxSTWv:3:CL:18:tag'
                        }
                    ],
                    'name': 'greenLevel',
                    'non_revoked': {
                        'to': 1532367957,
                        'from': 1532367957
                    }
                },
                '18_legalName_uuid': {
                    'restrictions': [
                        {
                            'cred_def_id': 'WgWxqztrNooG92RXvxSTWv:3:CL:18:tag'
                        }
                    ],
                    'name': 'legalName',
                    'non_revoked': {
                        'to': 1532367957,
                        'from': 1532367957
                    }
                },
                '15_id_uuid': {  # this specification will not show up in response: no cred def id restriction :-(
                    'name': 'id',
                    'non_revoked': {
                        'to': 1532367957,
                        'from': 1532367957
                    }
                }
            }
            'requested_predicates': {
            }
        }

    :return: nested dict mapping cred def id to name to proof request referent; e.g.,

    ::

        {
            'WgWxqztrNooG92RXvxSTWv:3:CL:18:tag': {
                'legalName': '18_legalName_uuid'
                'greenLevel': '18_greenLevel_uuid'
            }
        }

    """

    rv = {}
    for uuid, spec in proof_req['requested_attributes'].items():
        cd_id = None
        for restriction in spec.get('restrictions', []):
            cd_id = restriction.get('cred_def_id', None)
            if cd_id:
                break
        if not cd_id:
            continue
        if cd_id not in rv:  # cd_id of None is not OK
            rv[cd_id] = {}
        rv[cd_id][spec['name']] = uuid

    return rv


def proof_req_pred_referents(proof_req: dict) -> dict:
    """
    Given a proof request with all requested predicates having cred def id restrictions,
    return its predicate referents by cred def id and attribute, mapping a predicate and a limit.

    The returned structure can be useful in downstream processing to filter cred-infos for predicates.

    :param proof_req: proof request with all requested predicate specifications having cred def id restriction; e.g.,

    ::

        {
            'name': 'proof_req',
            'version': '0.0',
            'requested_attributes': {
                ...
            }
            'requested_predicates': {
                '194_highscore_GE_uuid': {
                    'name': 'highscore',
                    'p_type': '>=',
                    'p_value': '100000',
                    'restrictions': [
                        {
                            'cred_def_id': 'WgWxqztrNooG92RXvxSTWv:3:CL:194:tag'
                        }
                    ],
                    'non_revoked': {
                        ...
                    }
                },
                '194_level_GE_uuid': {
                    'name': 'level',
                    'p_type': '>=',
                    'p_value': '10',
                    'restrictions': [
                        {
                            'cred_def_id': 'WgWxqztrNooG92RXvxSTWv:3:CL:194:tag'
                        }
                    ],
                    'non_revoked': {
                        ...
                    }
                },
                '194_attempts_LE_uuid': {
                    'name': 'attempts',
                    'p_type': '<=',
                    'p_value': '3',
                    'restrictions': [
                        {
                            'cred_def_id': 'WgWxqztrNooG92RXvxSTWv:3:CL:194:tag'
                        }
                    ],
                    'non_revoked': {
                        ...
                    }
                },
                '198_employees_LT_uuid': {
                    'name': 'employees',
                    'p_type': '<',
                    'p_value': '100',
                    'restrictions': [
                        {
                            'cred_def_id': 'WgWxqztrNooG92RXvxSTWv:3:CL:198:tag'
                        }
                    ],
                    'non_revoked': {
                        ...
                    }
                },
                '198_employees_GE_uuid': {
                    'name': 'employees',
                    'p_type': '>=',
                    'p_value': '50',
                    'restrictions': [
                        {
                            'cred_def_id': 'WgWxqztrNooG92RXvxSTWv:3:CL:198:tag'
                        }
                    ],
                    'non_revoked': {
                        ...
                    }
                },
            }
        }

    :return: nested dict mapping cred def id to name to proof request referent to predicate and limit; e.g.,

    ::

        {
            'WgWxqztrNooG92RXvxSTWv:3:CL:194:tag': {
                'highscore': {
                    '194_level_GE_uuid': ['>=', 100000]
                },
                'level': {
                    '194_level_GE_uuid': ['>=', 10]
                },
                'attempts': {
                    '194_attempts_LE_uuid': ['<=', 3]
                }
            },
            'WgWxqztrNooG92RXvxSTWv:3:CL:198:tag': {
                'employees': {  # may have many preds per attr, but always 1 uuid and 1 relation per pred
                    '198_LT_employees_uuid': ['<=', 100]
                    '198_GE_employees_uuid': ['>=', 50]
                }
            }
        }

    """

    rv = {}
    for uuid, spec in proof_req['requested_predicates'].items():
        cd_id = None
        for restriction in spec.get('restrictions', []):
            cd_id = restriction.get('cred_def_id', None)
            if cd_id:
                break
        if not cd_id:
            continue
        if cd_id not in rv:  # cd_id of None is not OK
            rv[cd_id] = {}
        if spec['name'] not in rv[cd_id]:
            rv[cd_id][spec['name']] = {}
        rv[cd_id][spec['name']][uuid] = [spec['p_type'], Predicate.to_int(spec['p_value'])]

    return rv


def revoc_info(briefs: Union[dict, Sequence[dict]], filt: dict = None) -> dict:
    """
    Given a cred-brief, cred-info or sequence of either, return a dict mapping pairs
    (revocation registry identifier, credential revocation identifier)
    to attribute name: (raw) value dicts.

    If the caller includes a filter of attribute:value pairs, retain only matching attributes.

    :param briefs: cred-brief/cred-info, or sequence thereof
    :param filt: dict mapping attributes to values of interest; e.g.,

    ::

        {
            'legalName': 'Flan Nebula',
            'effectiveDate': '2018-01-01',
            'endDate': None
        }

    :return: dict mapping (rev_reg_id, cred_rev_id) pairs to (raw) attributes; e.g.,

    ::

        {
            ('LjgpST2rjsoxYegQDRm7EL:4:LjgpST2rjsoxYegQDRm7EL:3:CL:17:tag:CL_ACCUM:1', '2'): {
                'busId': '11121398',
                'effectiveDate': '2010-10-10',
                'endDate': '',
                'id': '1',
                'jurisdictionId': '1',
                'legalName': 'The Original House of Pies',
                'orgTypeId': '2'},
            ('LjgpST2rjsoxYegQDRm7EL:4:LjgpST2rjsoxYegQDRm7EL:3:CL:17:tag:CL_ACCUM:1', '3'): {
                'busId': '11133333',
                'effectiveDate': '2011-10-01',
                'endDate': '',
                'id': '2',
                'jurisdictionId': '1',
                'legalName': 'Planet Cake',
                'orgTypeId': '1'}
        }

    """

    rv = {}
    for brief in iter_briefs(briefs):
        cred_info = brief.get('cred_info', {}) or brief  # briefs could be cred-infos or cred-briefs
        (rr_id, cr_id) = (cred_info['rev_reg_id'], cred_info['cred_rev_id'])
        if (rr_id, cr_id) in rv or rr_id is None or cr_id is None:
            continue
        if not filt:
            rv[(rr_id, cr_id)] = cred_info['attrs']
            continue
        if ({attr: str(filt[attr]) for attr in filt}.items() <= cred_info['attrs'].items()):
            rv[(rr_id, cr_id)] = cred_info['attrs']

    return rv


def revealed_attrs(proof: dict) -> dict:
    """
    Fetch revealed attributes from input proof and return dict mapping credential definition identifiers
    to dicts, each dict mapping attribute names to (raw) values, for processing in further creds downstream.

    :param proof: indy-sdk proof as dict
    :return: dict mapping cred-ids to dicts, each mapping revealed attribute names to (raw) values
    """

    rv = {}
    for sub_index in range(len(proof['identifiers'])):
        cd_id = proof['identifiers'][sub_index]['cred_def_id']
        rv[cd_id] = ({  # uses von_anchor convention for uuid (referent) construction: will break on foreign anchor's
            '_'.join(uuid.split('_')[1:-1]): proof['requested_proof']['revealed_attrs'][uuid]['raw']
            for uuid in proof['requested_proof']['revealed_attrs']
            if proof['requested_proof']['revealed_attrs'][uuid]['sub_proof_index'] == sub_index})

    return rv



"""
Copyright 2017-2019 Government of Canada - Public Services and Procurement Canada - buyandsell.gc.ca

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""


import jsonschema

from von_anchor.error import JSONValidation


CONFIG_JSON_SCHEMA = {
    'pool': {
        '$schema': 'http://json-schema.org/draft-04/schema',
        'type': 'object',
        'properties': {
            'timeout': {
                'type': 'integer'
            },
            'extended_timeout': {
                'type': 'integer'
            },
            'preordered_nodes': {
                'type': 'array',
                'items': {
                    'type': 'string',
                    'uniqueItems': True
                }
            }
        },
        'additionalProperties': False
    },
    'holder-prover': {
        '$schema': 'http://json-schema.org/draft-04/schema',
        'type': 'object',
        'properties': {
            'parse-caches-on-open': {
                'type': 'boolean'
            },
            'archive-holder-prover-caches-on-close': {
                'type': 'boolean'
            }
        },
        'additionalProperties': True
    },
    'verifier': {
        '$schema': 'http://json-schema.org/draft-04/schema',
        'type': 'object',
        'properties': {
            'parse-caches-on-open': {
                'type': 'boolean'
            },
            'archive-verifier-caches-on-close': {
                'type': 'object',
                'properties': {
                    'schema_id': {
                        'type': 'array',
                        'items': {
                            'type': 'string',
                            'uniqueItems': True
                        }
                    },
                    'cred_def_id': {
                        'type': 'array',
                        'items': {
                            'type': 'string',
                            'uniqueItems': True
                        }
                    },
                    'rev_reg_id': {
                        'type': 'array',
                        'items': {
                            'type': 'string',
                            'uniqueItems': True
                        }
                    }
                }
            }
        },
        'additionalProperties': True
    },
    'org-hub': {
        '$schema': 'http://json-schema.org/draft-04/schema',
        'type': 'object',
        'properties': {
            'parse-caches-on-open': {
                'type': 'boolean'
            },
            'archive-holder-prover-caches-on-close': {
                'type': 'boolean'
            },
            'archive-verifier-caches-on-close': {
                'type': 'object',
                'properties': {
                    'schema_id': {
                        'type': 'array',
                        'items': {
                            'type': 'string',
                            'uniqueItems': True
                        }
                    },
                    'cred_def_id': {
                        'type': 'array',
                        'items': {
                            'type': 'string',
                            'uniqueItems': True
                        }
                    },
                    'rev_reg_id': {
                        'type': 'array',
                        'items': {
                            'type': 'string',
                            'uniqueItems': True
                        }
                    }
                }
            }
        },
        'additionalProperties': False
    }
}


def validate_config(key: str, config: dict) -> None:
    """
    Call jsonschema validation to raise JSONValidation on non-compliance or silently pass.

    :param key: validation schema key of interest
    :param config: configuration dict to validate
    """

    try:
        jsonschema.validate(config, CONFIG_JSON_SCHEMA[key])
    except jsonschema.ValidationError as x_valid:
        raise JSONValidation('JSON validation error on {} configuration: {}'.format(key, x_valid.message))
    except jsonschema.SchemaError as x_schema:
        raise JSONValidation('JSON schema error on {} specification: {}'.format(key, x_schema.message))



"""
Copyright 2017-2019 Government of Canada - Public Services and Procurement Canada - buyandsell.gc.ca

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""


import os
from os.path import dirname, join, realpath
from setuptools import setup, find_packages
from sys import stderr


pkg_name = 'von_anchor'
with open(join(dirname(__file__), 'VERSION.txt')) as fh_version:
    release = fh_version.read().strip()


def parse_requirements(filename):
    """
    Load requirements from a pip requirements file.

    :param filename: file name with requirements to parse
    """

    try:
        with open(filename) as fh_req:
            return [line.strip() for line in fh_req if line.strip() and not line.startswith('#')]
    except FileNotFoundError:
        print('File not found: {}'.format(realpath(filename)), file=stderr)
        raise


setup(
    name=pkg_name,
    version=release,
    packages=find_packages(exclude=['test']),
    description='VON anchors',
    entry_points={
        'console_scripts': [
            'von_anchor_setnym = von_anchor.op.setnym:main'
        ]
    },
    license='Apache Software License',
    author='PSPS-SPAC',
    author_email='stephen.klump@becker-carroll.com',
    url='https://github.com/PSPC-SPAC-buyandsell/{}'.format(pkg_name),
    download_url='https://github.com/PSPC-SPAC-buyandsell/{}/archive/{}.tar.gz'.format(pkg_name, release),
    keywords=['VON', 'SRI', 'anchor', 'The Org Book', 'indy'],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Software Development',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 3.5',
    ],
    python_requires='>=3.5',
    install_requires=parse_requirements('requirements.txt'),
)



"""
Copyright 2017-2019 Government of Canada - Public Services and Procurement Canada - buyandsell.gc.ca

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""


from .info import DIDInfo, EndpointInfo, KeyInfo, PairwiseInfo, pairwise_info2tags, storage_record2pairwise_info
from .record import StorageRecord
from .search import StorageRecordSearch
from .wallet import Wallet
from .manager import WalletManager



"""
Copyright 2017-2019 Government of Canada - Public Services and Procurement Canada - buyandsell.gc.ca

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""


from typing import List, Sequence, Union

from von_anchor.a2a.docutil import canon_did, canon_ref
from von_anchor.a2a.publickey import PublicKey


class Service:
    """
    Service specification to embed in DID document. Retains DIDs as raw values
    (orientated toward indy-facing operations), everything else as URIs
    (oriented toward W3C-facing operations).
    """

    def __init__(
            self,
            did: str,
            ident: str,
            typ: str,
            recip_keys: Union[Sequence, PublicKey],
            routing_keys: Union[Sequence, PublicKey],
            endpoint: str,
            priority: int = 0):
        """
        Retain service specification particulars. Raise BadIdentifier on bad input controller DID.

        :param did: DID of DID document embedding service, specified raw (operation converts to URI)
        :param ident: identifier for service
        :param typ: service type
        :param recip_keys: recipient key or keys
        :param routing_keys: routing key or keys
        :param endpoint: service endpoint
        :param priority: service priority
        """

        self._did = canon_did(did)
        self._id = canon_ref(self._did, ident, ';')
        self._type = typ
        self._recip_keys = (
            [recip_keys] if isinstance(recip_keys, PublicKey) else list(recip_keys) if recip_keys else None)
        self._routing_keys = (
            [routing_keys] if isinstance(routing_keys, PublicKey) else list(routing_keys) if routing_keys else None)
        self._endpoint = endpoint
        self._priority = priority

    @property
    def did(self) -> str:
        """
        Return DID.

        :return: DID
        """

        return self._did

    @property
    def id(self) -> str:
        """
        Return service identifier.

        :return: service identifier
        """

        return self._id

    @property
    def type(self) -> str:
        """
        Return service type.

        :return: service type
        """

        return self._type

    @property
    def recip_keys(self) -> List[PublicKey]:
        """
        Return recipient keys.

        :return: recip keys
        """

        return self._recip_keys

    @property
    def routing_keys(self) -> List[PublicKey]:
        """
        Return routing keys.

        :return: routing keys
        """

        return self._routing_keys

    @property
    def endpoint(self) -> str:
        """
        Return endpoint value.

        :return: endpoint value
        """

        return self._endpoint

    @property
    def priority(self) -> int:
        """
        Return priority value.

        :return: priority value
        """

        return self._priority

    def to_dict(self) -> dict:
        """
        Return dict representation of service to embed in DID document.
        """

        rv = {
            'id': self.id,
            'type': self.type,
            'priority': self.priority
        }
        if self.recip_keys:
            rv['recipientKeys'] = [k.value for k in self.recip_keys]
        if self.routing_keys:
            rv['routingKeys'] = [k.value for k in self.routing_keys]
        rv['serviceEndpoint'] = self.endpoint

        return rv
